"""
create_imgt_database.py
@author: leem
@date:   17 May 2017

Creates the IMGT database to be used for /data/machine/TCRDB_DATA/
"""
from html.parser import HTMLParser
import re, os, sys, urllib.request, urllib.parse, urllib.error

amino = "ACDEFGHIKLMNPQRSTVWY"

NUMBERINGS = {
    "Homsap": {
        "TRBV": [ "B%d" % (i+1) for i in range(108) ],
    },
    "Musmus": {
        "TRBV": [ "B%d" % (i+1) for i in range(108) ]

    }
}

NUMBERINGS["Homsap"]["TRAV"] = [ "A%d" % (i+1) for i in range(108) ]
NUMBERINGS["Homsap"]["TRGV"] = [ "G%d" % (i+1) for i in range(108) ]
NUMBERINGS["Homsap"]["TRDV"] = [ "D%d" % (i+1) for i in range(108) ]

trav_mus = []
for i in range(108):
    if i == 7 or i == 83:
        trav_mus.append( "A%d" % (i+1) )
        trav_mus.append( "A%dA"% (i+1) )
    else:
        trav_mus.append( "A%d" % (i+1) )

NUMBERINGS["Musmus"]["TRAV"] = trav_mus
NUMBERINGS["Musmus"]["TRGV"] = NUMBERINGS["Homsap"]["TRGV"]
NUMBERINGS["Musmus"]["TRDV"] = NUMBERINGS["Homsap"]["TRDV"]


SPECIES = {
    "Homsap":"Homo+sapiens",
    "Musmus":"Mus+musculus" # Mice for domaindisplay are "mus musculus"...
}
URL = "http://www.imgt.org/3Dstructure-DB/cgi/DomainDisplay.cgi?modif=1&species=%s&patternespece=&ReceptorType=TR&patternRecType=&groups=TR%s&patternGroup=&subgroups=any&patternSubgroup=&genes=any&patternGene=&alleles=any&patternAllele=&seldomtype=--&seldomnum=--&selaccnum=--&showallele=all&shownbseq=all&Show=Show+sequences&affesp=on&affge=on&afforig=on&affdom=on&affdomnum=on&afff=on&colorseq=1&difference=none&separespecie=none&separegroup=none&separesubgroup=none&separegene=none&separedomain=name&repeatheaddomain=true&tabChoice=Alignment"

# For J-regions we stick everything in
domains = ["V-BETA", "V-ALPHA", "V-GAMMA", "V-DELTA"] 

var_types = [
    "BV", "AV", "GV", "DV",
    "BJ", "AJ", "GJ", "DJ",  # For J-regions we stick everything in
]
DOMAIN_NAMES = {
    "BV": "V-BETA",
    "AV": "V-ALPHA",
    "GV": "V-GAMMA",
    "DV": "V-DELTA",

    "BC": "C-BETA",
    "AC": "C-ALPHA",
    "GC": "C-GAMMA",
    "DC": "C-DELTA",
}

MASTER_DICT = {}


class IMGTParser(HTMLParser):
    currenttag = None
    currentnamedent = None

    def handle_starttag(self, tag, attrs):
        self.currenttag = tag
        for a in attrs:
            if a[1] == "allelename" and tag == "td":
                self.allelename = True
            if tag == "a" and "ligmdb" in a[1]:
            	self.accesion = True
        
    def handle_endtag(self, tag):
        self.currenttag = None

    def handle_data(self, data):
        split = data.split("\n")

        # Species names are under em tags        
        if self.currenttag == "em":
            split = split[0]
            self.species = split
            
            # If species isn't human or mouse, skip
            if self.species not in SPECIES:
                self.species = None

            # If species is currently not logged, then put them in dict
            if self.species in SPECIES and self.species not in self._data:
                self._data[self.species] = dict()
        
        # If we get an em tag with no species, skip
        if not self.species:
            return;
        
        if self.currenttag == "td":
            split = split[0].strip()
            if split == "P" or split == "ORF":
                MASTER_DICT[ self.allele ] = True
                self.pseudo = True
                return
            elif split == "F":
                MASTER_DICT[ self.allele ] = False
                self.pseudo = False

        # a tags are unusual; they contain allele names (in the first instance) but contain other accesion stuff afterward
        # Get the first a tag assume as allele name, then skip
        if self.currenttag == "a" and self.species in SPECIES:
            split = split[0]

            if split and self.allelename:
                self.allele = split
                self.allelename = False

                if self.allele not in self._data[self.species]:
                    self._data[self.species][self.allele] = {}

            if split and self.accesion:
                self.accnum = split
                self.accesion = False
                self._data[self.species][self.allele][self.accnum] = ""

        # Once we've established a domain, then get the sequence
        if self.accnum:
            if (self.currenttag == "pre" or self.currenttag == "span" or self.currenttag == None or self.currenttag == "b"):
                split = split[0].strip()
                self._data[self.species][self.allele][self.accnum] += split

        # Set the prevtag variable to what it is currently; this is for sequences that are flanked by IMGT's N-linked glycosylation sites
        self.prevtag = self.currenttag
        self.prevallele = self.allele

    def rip_sequences(self,htmlstring):
        """
        Method for this subclass that automates the return of data
        """
        self.reset()
        self._data = dict()
        self.currenttag = None
        self.currentnamedent = None

        self.pre = False

        self.prevtag = None
        self.species = None
        self.allele = None
        self.prevallele = None
        self.allelename = False
        self.accesion = False
        self.accnum = None
        self.pseudo = False

        self.feed(htmlstring)

        return self._data

class IMGT_J_Parser(IMGTParser):
    currenttag = None
    currentnamedent = None

    def handle_starttag(self, tag, attrs):
        self.currenttag = tag
        for a in attrs:
            if a[1] == "allelename" and tag == "td":
                self.allelename = True
            if tag == "a" and "ligmdb" in a[1]:
                self.accesion = True
        
    def handle_endtag(self, tag):
        self.currenttag = None

    def handle_data(self, data):
        split = data.split("\n")

        # Species names are under em tags        
        if self.currenttag == "em":
            split = split[0]
            self.species = split
            
            # If species isn't human or mouse, skip
            if self.species not in SPECIES:
                self.species = None

            # If species is currently not logged, then put them in dict
            if self.species in SPECIES and self.species not in self._data:
                self._data[self.species] = dict()
        
        # If we get an em tag with no species, skip
        if not self.species:
            return;

        # Get the allele name which is in td
        if self.currenttag == "td":
            split = split[0].strip()

            if split == "P" or split == "ORF":
                MASTER_DICT[ self.allele ] = True
                self.pseudo = True
                return
            elif split == "F":
                MASTER_DICT[ self.allele ] = False
                self.pseudo = False

            if split and self.allelename:
                self.allele = split
                self.allelename = False
                if self.allele and self.allele not in self._data[self.species]:
                    self._data[self.species][self.allele] = {}

        # a tags are unusual; they contain allele names (in the first instance) but contain other accesion stuff afterward
        # Get the first a tag assume as allele name, then skip
        if self.currenttag == "a" and self.species in SPECIES:
            split = split[0]

            if split and self.accesion:
                self.accnum = split
                self.accesion = False
                self._data[self.species][self.allele][self.accnum] = ""
        
        # Once we've established an accnum, then get the sequence
        if self.accnum:            
            if not self.pseudo and (self.currenttag == "pre" or self.currenttag == "span" or self.currenttag == None or self.currenttag == "b"):
                split = split[0].strip()
                self._data[self.species][self.allele][self.accnum] += split

        # Set the prevtag variable to what it is currently; this is for sequences that are flanked by IMGT's N-linked glycosylation sites
        self.prevtag = self.currenttag
        self.prevallele = self.allele

    def rip_sequences(self,htmlstring):
        """
        Method for this subclass that automates the return of data
        """
        self.reset()
        self._data = dict()
        self.currenttag = None
        self.currentnamedent = None

        self.pre = False
        self.pseudo = False

        self.prevtag = None
        self.species = None
        self.allele = None
        self.prevallele = None
        self.allelename = False
        self.accesion = False
        self.accnum = None

        self.feed(htmlstring)

        return self._data

class IMGT_C_Parser(IMGT_J_Parser):
    currenttag = None
    currentnamedent = None

    def handle_starttag(self, tag, attrs):
        self.currenttag = tag
        for a in attrs:
            if a[1] == "allelename" and tag == "td":
                self.allelename = True
            if tag == "a" and "ligmdb" in a[1]:
                self.accesion = True
        
    def handle_endtag(self, tag):
        self.currenttag = None

    def handle_data(self, data):
        split = data.split("\n")

        # Species names are under em tags        
        if self.currenttag == "em":
            split = split[0]
            self.species = split
            
            # If species isn't human or mouse, skip
            if self.species not in SPECIES:
                self.species = None

            # If species is currently not logged, then put them in dict
            if self.species in SPECIES and self.species not in self._data:
                self._data[self.species] = dict()
        
        # If we get an em tag with no species, skip
        if not self.species:
            return;

        # Get the allele name which is in td
        if self.currenttag == "a":
            split = split[0].strip()

            if split == "P" or split == "ORF":
                print(self.species, self.allele)
                self.pseudo = True
                return
            elif split == "F" or split == "(F)":
                self.pseudo = False

            if split and self.allelename:
                self.allele = split
                self.allelename = False
                if self.allele and self.allele not in self._data[self.species]:
                    self._data[self.species][self.allele] = {}

        # a tags are unusual; they contain allele names (in the first instance) but contain other accesion stuff afterward
        # Get the first a tag assume as allele name, then skip
        if self.currenttag == "a" and self.species in SPECIES and self.allele and self.accesion:
            self.accnum = split
            print(self.accnum)
            self.accesion = False
            self._data[self.species][self.allele][self.accnum] = ""

        # Once we've established an accnum, then get the sequence
        if self.accnum and self.allele:            
            if not self.pseudo and (self.currenttag == "pre" or self.currenttag == "span" or self.currenttag == None or self.currenttag == "b"):
                split = split[0].strip()
                self._data[self.species][self.allele][self.accnum] += split

        # Set the prevtag variable to what it is currently; this is for sequences that are flanked by IMGT's N-linked glycosylation sites
        self.prevtag = self.currenttag
        self.prevallele = self.allele

    def rip_sequences(self,htmlstring):
        """
        Method for this subclass that automates the return of data
        """
        self.reset()
        self._data = dict()
        self.currenttag = None
        self.currentnamedent = None

        self.pre = False
        self.pseudo = False

        self.prevtag = None
        self.species = None
        self.allele = None
        self.prevallele = None
        self.allelename = False
        self.accesion = False
        self.accnum = None

        self.feed(htmlstring)

        return self._data


class IMGTDatabase(object):
    """
        Create an IMGT database based on the results from DomainDisplay
    """
    def __init__(self, parser, var_type):
        self.data = parser._data
        self.species = list(self.data.keys())[0]
        self.vartype   = var_type
        self.cleanup()
        self.numbering = NUMBERINGS

    def cleanup(self):
        """
            Clean up the sequences that are parsed from IMGT
        """
        for allele in self.data[self.species]:
            for accnum in self.data[self.species][allele]:
                self.data[self.species][allele][accnum] = self.data[self.species][allele][accnum].replace(" ", "") # Get rid of spaces
                self.data[self.species][allele][accnum] = self.data[self.species][allele][accnum].replace(".", "-") # Get rid of dots as dashes
                seqlen = len(self.data[self.species][allele][accnum])

                if self.species == "Musmus" and "A" in self.vartype:
                    if seqlen < 110 and "V" in self.vartype:
                        self.data[self.species][allele][accnum] += "-" * (110-seqlen)
                    elif seqlen >= 110 and "V" in self.vartype:
                        self.data[self.species][allele][accnum] = self.data[self.species][allele][accnum][:110] # keep an equivalent length of 110
                else:
                    if seqlen < 108 and "V" in self.vartype:
                        self.data[self.species][allele][accnum] += "-" * (108-seqlen)
                    elif seqlen >= 108 and "V" in self.vartype:
                        self.data[self.species][allele][accnum] = self.data[self.species][allele][accnum][:108] # keep an equivalent length of 108

                if "C" in self.vartype:
                    self.data[self.species][allele][accnum] = self.data[self.species][allele][accnum].replace("-", "") # Get rid of dashes completely in the C domain

    def save(self, out = sys.stdout):

        header = "Species\tGene\tAllele\tAccNum\tDomain label\tGroup\tSubgroup\t"
        if "J" in self.vartype:
            header += "Sequence"
            domain_name = "J-REGION"
        elif "C" in self.vartype:
            header += "Sequence"
            domain_name = DOMAIN_NAMES[self.vartype]
        else:
            header += "\t".join(NUMBERINGS[self.species]["TR%s" % self.vartype])
            domain_name = DOMAIN_NAMES[self.vartype]

        common_name = {
            "Homsap": "HOMO SAPIENS",
            "Musmus": "MUS MUSCULUS"
        }

        species_name = common_name[self.species]

        dbstring = header + "\n"

        for allele in self.data[self.species]:
            # Don't include ORFs and P genes
            if allele in MASTER_DICT and MASTER_DICT[allele]:
                continue
            if allele.split("*")[1] != "01":
                continue

            gene = allele.split("*")[0]

            if "C" in gene:
                subgroup = "?"
            elif "/" not in gene:
                try:
                    subgroup = re.findall(r'TR\w*\d', gene)[0]
                except IndexError:
                    subgroup = re.findall(r'TR\w*', gene)[0] # TRGVA case, human

            elif "/" in gene and "A" in self.vartype:
                subgroup = re.findall(r'TR\w*\d', gene)[0]
            elif "/" in gene and "D" in self.vartype:
                subgroup = re.findall(r'TRD\d', gene)[0] # Cases where D/A genes can be used

            for accnum in self.data[self.species][allele]:

                # Blank sequences; skip
                if not self.data[self.species][allele][accnum]:
                    continue

                if self.vartype == "BC" or self.vartype == "GC":
                    domain_name = DOMAIN_NAMES[self.vartype] + "-%s" % gene[-1]

                dbstring += "%s\t%s\t%s\t%s\t%s\t%s\t%s\t" % (species_name, gene, allele, accnum, domain_name, gene[:4], subgroup)

                if "J" in gene or "C" in gene:
                    dbstring += self.data[self.species][allele][accnum]
                else:
                    dbstring += "\t".join(list(self.data[self.species][allele][accnum]))

                dbstring += "\n"

        out.write(dbstring)
        self.dbstring = dbstring

import sys

the_species = sys.argv[1] # e.g. Homo+sapiens
the_vartype = sys.argv[2] # e.g. AC, AV, AJ

print('Querying IMGT...')
if not os.path.isfile("%s%s.html" % (the_species, the_vartype)):
    f = urllib.request.urlopen(URL % (the_species, the_vartype))

    with open("%s%s.html" % (the_species, the_vartype), 'w') as outf:
        for line in f:
            print(line.strip(), file=outf)

f = open("%s%s.html" % (the_species, the_vartype))

print('Parsing IMGT HTML...')

if "J" in the_vartype:
    p = IMGT_J_Parser()
elif "C" in the_vartype:
    p = IMGT_C_Parser()
else:
    p = IMGTParser()
p.rip_sequences(f.read())

print('Saving Database...')
db = IMGTDatabase(p, the_vartype)
db.save(open("%s_%s.dat" % (the_species, the_vartype), 'w'))
if the_vartype == "GC" or the_vartype == "BC":
    sys.stderr.write("Manually check the dat file for %s_%s.dat\n" % (the_species, the_vartype))
