#    ANARCI - Antibody Numbering and Antigen Receptor ClassIfication
#    Copyright (C) 2015 Oxford Protein Informatics Group (OPIG)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.#
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Module containing functions to convert hmm alignment to a numbering scheme. 
@author: dunbar
@editor: leem

Currently implemented
For TR's and MHC's
IMGT

---------------------------------------------------------------------------------------------------------------------
Functions are written to a template:

There are 128 match states in the HMMs (these are the IMGT states). The alignment to these states must be converted to
correspond to the scheme of choice. 

We define:
  - a state string consisting of 'X' and 'I' where:
    X  means that for the state there is an equivalent position in the numbering scheme.
    I  means that for the state there is not an equivalent position in the numbering scheme. It should therefore be 
       considered as an insertion in the scheme.
       
  - a region string consisting of characters (integers in the currently implemented schemes). Each character 
corresponds to a contiguous region. Therefore each state can be assigned a region according to the scheme. 

  - a mapping between region characters and region indices as a dictionary. e.g. the first region character maps
to 0, second to 1 ...

  - a dictionary containing the difference between state number (imgt) and scheme number at the *beginning* of 
each region using the region indices as keys and the difference as values.

  - the number of regions defined

  - a list for which delete states should not be included in the numbering (typically those for the cdrs). This
will allow the length of the region to be the number of residues found instead of the number of possible states plus
insertions.

 
This all goes into the _number_regions function along with the sequence and the state_vector (the alignment from the
HMM).

_number regions will then divide the aligned part of the sequence into as many regions as defined above. Within each 
region it will give a numbering according to the input parameters. A list of lists will be returned containing the 
numbered sequence for each region.

Some of the regions will not be numbered correctly according to the scheme. For example the insertions for the CDRs
will not necessarily be on the correct residue. For each different scheme these regions are then modified (see code
for implementation) 

Finally the full numbered sequence is compiled and returned to the calling function.
---------------------------------------------------------------------------------------------------------------------

Other schemes can be implemented following the template above. 

'''

# Alphabet used for insertion (last (-1th) is a blank space for no insertion)
alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZ "

# Blosum62 matrix. Used in some annotation methods to recognise pre-defined motifs
blosum62 = {('B', 'N'): 3, ('W', 'L'): -2, ('G', 'G'): 6, ('X', 'S'): 0, ('X', 'D'): -1, ('K', 'G'): -2, ('S', 'E'): 0, ('X', 'M'): -1, ('Y', 'E'): -2, ('W', 'R'): -3, ('I', 'R'): -3, ('X', 'Z'): -1, ('H', 'E'): 0, ('V', 'M'): 1, ('N', 'R'): 0, ('I', 'D'): -3, ('F', 'D'): -3, ('W', 'C'): -2, ('N', 'A'): -2, ('W', 'Q'): -2, ('L', 'Q'): -2, ('S', 'N'): 1, ('Z', 'K'): 1, ('V', 'N'): -3, ('Q', 'N'): 0, ('M', 'K'): -1, ('V', 'H'): -3, ('G', 'E'): -2, ('S', 'L'): -2, ('P', 'R'): -2, ('D', 'A'): -2, ('S', 'C'): -1, ('E', 'D'): 2, ('Y', 'G'): -3, ('W', 'P'): -4, ('X', 'X'): -1, ('Z', 'L'): -3, ('Q', 'A'): -1, ('V', 'Y'): -1, ('W', 'A'): -3, ('G', 'D'): -1, ('X', 'P'): -2, ('K', 'D'): -1, ('T', 'N'): 0, ('Y', 'F'): 3, ('W', 'W'): 11, ('Z', 'M'): -1, ('L', 'D'): -4, ('M', 'R'): -1, ('Y', 'K'): -2, ('F', 'E'): -3, ('M', 'E'): -2, ('S', 'S'): 4, ('X', 'C'): -2, ('Y', 'L'): -1, ('H', 'R'): 0, ('P', 'P'): 7, ('K', 'C'): -3, ('S', 'A'): 1, ('P', 'I'): -3, ('Q', 'Q'): 5, ('L', 'I'): 2, ('P', 'F'): -4, ('B', 'A'): -2, ('Z', 'N'): 0, ('M', 'Q'): 0, ('V', 'I'): 3, ('Q', 'C'): -3, ('I', 'H'): -3, ('Z', 'D'): 1, ('Z', 'P'): -1, ('Y', 'W'): 2, ('T', 'G'): -2, ('B', 'P'): -2, ('P', 'A'): -1, ('C', 'D'): -3, ('Y', 'H'): 2, ('X', 'V'): -1, ('B', 'B'): 4, ('Z', 'F'): -3, ('M', 'L'): 2, ('F', 'G'): -3, ('S', 'M'): -1, ('M', 'G'): -3, ('Z', 'Q'): 3, ('S', 'Q'): 0, ('X', 'A'): 0, ('V', 'T'): 0, ('W', 'F'): 1, ('S', 'H'): -1, ('X', 'N'): -1, ('B', 'Q'): 0, ('K', 'A'): -1, ('I', 'Q'): -3, ('X', 'W'): -2, ('N', 'N'): 6, ('W', 'T'): -2, ('P', 'D'): -1, ('B', 'C'): -3, ('I', 'C'): -1, ('V', 'K'): -2, ('X', 'Y'): -1, ('K', 'R'): 2, ('Z', 'R'): 0, ('W', 'E'): -3, ('T', 'E'): -1, ('B', 'R'): -1, ('L', 'R'): -2, ('Q', 'R'): 1, ('X', 'F'): -1, ('T', 'S'): 1, ('B', 'D'): 4, ('Z', 'A'): -1, ('M', 'N'): -2, ('V', 'D'): -3, ('F', 'A'): -2, ('X', 'E'): -1, ('F', 'H'): -1, ('M', 'A'): -1, ('K', 'Q'): 1, ('Z', 'S'): 0, ('X', 'G'): -1, ('V', 'V'): 4, ('W', 'D'): -4, ('X', 'H'): -1, ('S', 'F'): -2, ('X', 'L'): -1, ('B', 'S'): 0, ('S', 'G'): 0, ('P', 'M'): -2, ('Y', 'M'): -1, ('H', 'D'): -1, ('B', 'E'): 1, ('Z', 'B'): 1, ('I', 'E'): -3, ('V', 'E'): -2, ('X', 'T'): 0, ('X', 'R'): -1, ('R', 'R'): 5, ('Z', 'T'): -1, ('Y', 'D'): -3, ('V', 'W'): -3, ('F', 'L'): 0, ('T', 'C'): -1, ('X', 'Q'): -1, ('B', 'T'): -1, ('K', 'N'): 0, ('T', 'H'): -2, ('Y', 'I'): -1, ('F', 'Q'): -3, ('T', 'I'): -1, ('T', 'Q'): -1, ('P', 'L'): -3, ('R', 'A'): -1, ('B', 'F'): -3, ('Z', 'C'): -3, ('M', 'H'): -2, ('V', 'F'): -1, ('F', 'C'): -2, ('L', 'L'): 4, ('M', 'C'): -1, ('C', 'R'): -3, ('D', 'D'): 6, ('E', 'R'): 0, ('V', 'P'): -2, ('S', 'D'): 0, ('E', 'E'): 5, ('W', 'G'): -2, ('P', 'C'): -3, ('F', 'R'): -3, ('B', 'G'): -1, ('C', 'C'): 9, ('I', 'G'): -4, ('V', 'G'): -3, ('W', 'K'): -3, ('G', 'N'): 0, ('I', 'N'): -3, ('Z', 'V'): -2, ('A', 'A'): 4, ('V', 'Q'): -2, ('F', 'K'): -3, ('T', 'A'): 0, ('B', 'V'): -3, ('K', 'L'): -2, ('L', 'N'): -3, ('Y', 'N'): -2, ('F', 'F'): 6, ('L', 'G'): -4, ('B', 'H'): 0, ('Z', 'E'): 4, ('Q', 'D'): 0, ('X', 'B'): -1, ('Z', 'W'): -3, ('S', 'K'): 0, ('X', 'K'): -1, ('V', 'R'): -3, ('K', 'E'): 1, ('I', 'A'): -1, ('P', 'H'): -2, ('B', 'W'): -4, ('K', 'K'): 5, ('H', 'C'): -3, ('E', 'N'): 0, ('Y', 'Q'): -1, ('H', 'H'): 8, ('B', 'I'): -3, ('C', 'A'): 0, ('I', 'I'): 4, ('V', 'A'): 0, ('W', 'I'): -3, ('T', 'F'): -2, ('V', 'S'): -2, ('T', 'T'): 5, ('F', 'M'): 0, ('L', 'E'): -3, ('M', 'M'): 5, ('Z', 'G'): -2, ('D', 'R'): -2, ('M', 'D'): -3, ('W', 'H'): -2, ('G', 'C'): -3, ('S', 'R'): -1, ('S', 'I'): -2, ('P', 'Q'): -1, ('Y', 'A'): -2, ('X', 'I'): -1, ('E', 'A'): -1, ('B', 'Y'): -3, ('K', 'I'): -3, ('H', 'A'): -2, ('P', 'G'): -2, ('F', 'N'): -3, ('H', 'N'): 1, ('B', 'K'): 0, ('V', 'C'): -1, ('T', 'L'): -1, ('P', 'K'): -1, ('W', 'S'): -3, ('T', 'D'): -1, ('T', 'M'): -1, ('P', 'N'): -2, ('K', 'H'): -1, ('T', 'R'): -1, ('Y', 'R'): -2, ('L', 'C'): -1, ('B', 'L'): -4, ('Z', 'Y'): -2, ('W', 'N'): -4, ('G', 'A'): 0, ('S', 'P'): -1, ('E', 'Q'): 2, ('C', 'N'): -3, ('H', 'Q'): 0, ('D', 'N'): 1, ('Y', 'C'): -2, ('L', 'H'): -3, ('E', 'C'): -4, ('Z', 'H'): 0, ('H', 'G'): -2, ('P', 'E'): -1, ('Y', 'S'): -2, ('G', 'R'): -2, ('B', 'M'): -3, ('Z', 'Z'): 4, ('W', 'M'): -1, ('Y', 'T'): -2, ('Y', 'P'): -3, ('Y', 'Y'): 7, ('T', 'K'): -1, ('Z', 'I'): -3, ('T', 'P'): -1, ('V', 'L'): 1, ('F', 'I'): 0, ('G', 'Q'): -2, ('L', 'A'): -1, ('M', 'I'): 1}

# General function to give annotations for regions that have direct mappings onto the hmm alignment (imgt states)

def _number_regions(sequence, state_vector, state_string , region_string,  region_index_dict, rels, n_regions, exclude_deletions):
    """
    General function to number a sequence and divide it into different regions  
    
    @param sequence: The sequence string
    @param state_vector: The list of states from the aligned hmm
    @param state_string: A string of states for the scheme relative to IMGT (this is X for a direct equivalence, I if needs to be treated as insertion)
    @param region_string: A string of characters that indicate which hmm states are in each regions for this scheme (i.e. how should the sequence be divided up)
    @param region_index_dict: A dictionary converting the characters in region string to an index of the regions. 
    @param rels: The difference of the numbering integer at the *start* of each region
    @param n_regions: The number of regions
    @param exclude_deletions: A list of region indices for which deletion states should not be included. Typically the CDRs.
    
    @return: A list of lists where each region has been numbered according to the scheme. Some regions will need renumbering. This should be taken care of after the function called.
    
    """

    _regions = [ [] for _ in range(n_regions) ]
    
    # Initialise the insertion index (-1 is a blank space) and the previous state.
    insertion = -1
    previous_state_id = 1 
    start_index, end_index  = None, None
    
    region = None
    # Iterate over the aligned state vector
    for (state_id, state_type ), si in state_vector:

        # Retrieve the region index
        if state_type != "i" or region is None: # BUG_FIX - JD 9/4/15 - do not allow a new region to start as an insertion.
            region = region_index_dict[region_string[state_id-1]] 
        
        # Check the state_types
        if state_type == "m": # It is a match
            
            # Check whether this position is in the scheme as an independent state
            if state_string[state_id-1]=="I": # No, it should be treated as an insertion 
                insertion +=1 # Increment the insertion annotation index
                rels[region] -= 1 # Update the relative numbering from the imgt states
            else: # Yes 
                insertion = -1 # Reset the insertions 
            
            # Add the numbering annotation to the appropriate region list            
            _regions[region].append( ( (state_id + rels[region], alphabet[insertion] ), sequence[si]  ) )
            previous_state_id = state_id # Record the previous state ID
            if start_index is None:
                start_index = si
            end_index = si
            
        elif state_type == "i": # It is an insertion
            insertion +=1 # Increment the insertion annotation index
            
            # Add the numbering annotation to the appropriate region list
            _regions[region].append( ( (previous_state_id + rels[region], alphabet[insertion]), sequence[si]  ) )
            if start_index is None:
                start_index = si
            end_index = si

        else: # It is a deletion
            
            # Check whether this position is in the scheme as an independent state
            if state_string[state_id-1]=="I": # No, therefore irrelevant to the scheme.
                rels[region] -= 1 # Update the relative numbering from the imgt states
                continue 
            
            # Check whether we should add the delete state to the numbering (don't if it needs ammending)  
            if region not in exclude_deletions: # Don't add deletions
                _regions[region].append( ( (state_id + rels[region], " "), "-"  ) )
            
            insertion = -1 # Reset the insertions
            previous_state_id = state_id # Record the previous state ID, should not be needed (no delete to insert state transition)
        
        assert insertion < 25, "Too many insertions for numbering scheme to handle" # We ran out of letters. Cows will not be happy :( 
            
    return _regions, start_index, end_index


# Functions to perform the numbering and the corrections for each of the implemented schemes.
# These have been written fairly verbosely so that the template of how to generate a function for a new scheme is more clear.
# They have two stages: Perform the mapping between imgt and the scheme; Renumber those regions that do not map nicely onto imgt (e.g. CDR insertions)
    

########
# IMGT #
########

def apply_loop_numbering(prev_numbering, previous_state_id, loop_numbering):
    new_numbering = []
    si = 0
    
    for annotation in loop_numbering:
        if annotation is None:
            new_numbering.append( ((previous_state_id+1, " "), "-"   ) )
            previous_state_id += 1
        else:
            new_numbering.append( (annotation, prev_numbering[si][1] ) )
            previous_state_id = annotation[0]
            si += 1

    return new_numbering

def number_b2m(state_vector, sequence):
    """
    Apply the IMGT numbering scheme for 
    There are 114 states in the state vector for numbering B2M domains.
    Numbering the B2M is slightly tricky because there are FOUR regions; the BC loop actually skips a few numbers, and residue 45 is followed by 77.

    region 1: Astrand + AB loop + Bstrand
    region 2: BC loop (needs its own numbering)
    region 3: C strand up to and including residue 76
    region 4: D strand
    region 5: DE loop (which needs its own numbering)
    region 6: E, F, G strands with EF, FG loops
    """
                    # "|........|....|123|......|..||........||.|...|1234567|......|12345677654321|...|......|12|......||...........||..."
    state_string    = 'XXXXXXXXXXXXXXXIIIXXXXXXXXXXXXXXXXXXXXXXXXXXXXIIIIIIIXXXXXXXXIIIIIIIIIIIIIIXXXXXXXXXXXXIIXXXXXXXXXXXXXXXXXXXXXXXXX'
    region_string   = "111111111111111111111111111112222222222333333333333334444444455555555555555666666666666666666666666666666666666666"
    n_regions = 6 # Numbering jumps after residue 45; hence we have two regions to number
    exclude_deletions = [1]
    region_index_dict = {"1":0,"2":1, "3": 2, "4": 3, "5": 4, "6": 5}
    rels              =  {0:0, 1:0, 2:0, 3:0, 4:0, 5:0, 6:0}

    _regions, startindex, endindex = _number_regions(sequence, state_vector, state_string , region_string,  region_index_dict, rels, n_regions, exclude_deletions)

    # For edge cases, such as 4x6d; the sequence is fragmented, assert that the cysteine is always residue 23;
    # since this domain is pretty consistent, this is a hack that allows us to get a minimalistic sequence of the B2M
    if len(_regions[0]) < 10:
        blank  = []
        dictio = dict([ (y[1], y[0]) for y in _regions[0] ])

        if dictio["C"] != (23, ' '):
            cys_index = _regions[0].index((dictio["C"], "C"))
            newidx    = max(0, cys_index-3)
            start     = 23 - newidx # start from at most 3 positions behind the cysteine for numbering.
            for j, item in enumerate(_regions[0][newidx:]):
                blank.append(((start+j, ' '), item[1]))
        _regions[0] = blank

    bcLoop_numbering  = get_bc_loop(len(_regions[1]))
    previous_state_id = 27
    bc_numbering      = apply_loop_numbering(_regions[1], previous_state_id, bcLoop_numbering)

    # Region following the BC loop has been incremented by 1
    _regions[2] = [ ((_[0][0]-1, _[0][1]),  _[1])  for _ in _regions[2] ]

    # The D strand jumps in numbering, and adding 23 seems to do the trick (since there are insertions between 45 and 77)
    _regions[3] = [ ((_[0][0]+23, _[0][1]), _[1]) for _ in _regions[3]  ] 

    # 
    deLoop_numbering = get_de_loop(len(_regions[4]))
    de_numbering, si = [], 0
    for annotation in deLoop_numbering:
        if annotation is not None:
            de_numbering.append((annotation, _regions[4][si][1]))
            si += 1

    _regions[5] = [ ((_[0][0]+9, _[0][1]), _[1]) for _ in _regions[5]  ]

    return _regions[0] + bc_numbering + _regions[2] + _regions[3] + de_numbering + _regions[5], startindex, endindex

def number_mhc(state_vector, sequence):
    """
    Apply the IMGT numbering scheme for numbering the G domains of MHCs.
    MHCs have 106 states including insertions; check the DomainDisplay tool of IMGT.

    There are 106 states in the HMM of MHC molecules, and we treat X's as matches and I's as insertions.
    We don't need to do any separate re-numbering. This is a simple conversion of the state vector into the IMGT numbering scheme for MHCs.
    """
    state_string = 'XXXXXXXIXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXIIIIIIIXXXXXIIIXXXXXXXIIXXXXXXXXXXXIXXXXXXXXXXXXXXXXXXXX'
    region_string = "1"*106

    n_regions, exclude_deletions = 1, [1]
    region_index_dict = {"1":0}
    rels              =  {0:0} 

    _regions, startindex, endindex = _number_regions(sequence, state_vector, state_string , region_string,  region_index_dict, rels, n_regions, exclude_deletions)

    # Get the 0th item in _regions as _regions is a nested-list otherwise for MHC numbering. #10/5/17
    return _regions[0], startindex, endindex

def number_cd1(state_vector, sequence):
    """
    Apply the IMGT numbering scheme for numbering the G domains of CD1.
    CD1 has 104 states including insertions; check the DomainDisplay tool of IMGT.

    There are 104 states in the HMM of CD1 molecules, and we treat X's as matches and I's as insertions.
    """
    state_string = 'XXXXXXXIXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXIIIIIIIXXXXXIXXXXXXXIIXXXXXXXXXXXIXXXXXXXXXXXXXXXXXXXX'
    region_string = "1"*104

    n_regions, exclude_deletions = 1, [1]
    region_index_dict = {"1":0}
    rels              =  {0:0} 

    _regions, startindex, endindex = _number_regions(sequence, state_vector, state_string , region_string,  region_index_dict, rels, n_regions, exclude_deletions)

    # Get the 0th item in _regions as _regions is a nested-list otherwise for MHC numbering. #10/5/17
    return _regions[0], startindex, endindex

def number_imgt(state_vector, sequence):
    """    
    Apply the IMGT numbering scheme for heavy or light chains
    
    Rules should be implemented using two strings - the state string and the region string. 

    There are 128 states in the HMMs. Treat X as a direct match in IMGT scheme, I is an insertion. (All X's for IMGT)
    XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXX XXXXXXXXXXX
    11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111 2222222222222 33333333333

    XXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXX XXXXXXXXXXXXXXXXX XXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXX XXXXXXXXXXX
    11111111111111111111111111 222222222222 33333333333333333 4444444444 555555555555555555555555555555555555555 6666666666666 77777777777
    Regions - (N.B These do not match up with any particular definition of CDR)
    1. All positions before CDR1
    2. CDR1 positions
    3. Positions between CDR1/2
    4. CDR2 positions
    5. Positions between CDR2/3
    6. CDR positions 105 (inc) to 118 (exc)
    7. Positions after CDR3    
    
    """
    
    # Set up the numbering 
 
    # State string - 'X' means the imgt position exists in the scheme. 'I' means that it should be treated as an insertion of the previous number
    state_string =  'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'
                    
    # Region string - regions that should be treated separately in putting the numbering together 
    # region_string = '11111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111222222222222233333333333'
    region_string = '11111111111111111111111111222222222222333333333333333334444444444555555555555555555555555555555555555555666666666666677777777777'

    region_index_dict = {
                         "1":0,
                         "2":1,
                         "3":2,
                         "4":3,
                         "5":4,
                         "6":5,
                         "7":6
                         }
    
    # Define how the scheme's numbering differs from IMGT at the start of each region. 
    # This is updated in the loop below
    rels              =  {0:0, 
                          1:0,
                          2:0,
                          3:0,
                          4:0,
                          5:0,
                          6:0,
                          7:0
                          }
    
    n_regions = 7

    exclude_deletions = [1, 3, 5]
    
    _regions, startindex, endindex = _number_regions(sequence, state_vector, state_string , region_string,  region_index_dict, rels, n_regions, exclude_deletions)

    ###############
    # Renumbering #
    ###############

    _numbering = [ _regions[0], # Fw1
                   [],          # CDR1
                   _regions[2], # Fw2
                   [],          # CDR2
                   _regions[4], # Fw3
                   [],          # CDR3
                   _regions[6], # Fw4

                 ]

    # All CDRS require renumbering
    # They are numbered in a specific way for each CDRlength - i.e. we gap from the sides. Insertions are at 111 and 112 symetrically.

    # Get the sequence from _regions -- this is sometimes not aligned properly by HMMER so we need to get the sequence manually
    cdr1seq    = "".join([ x[1] for x in _regions[1] if x[1] != "-" ])
    cdr1length = len(cdr1seq)
    si = 0
    # Get re-annotated CDR1; CDR1 starts from 27-38. Thus range from 27 (inc.) to 39 (exc.)
    # has a theoretical length of 12.
    cdr1_annotations = get_imgt_cdr(cdr1length, 12, 27, 39)
    prev_state = 26
    for ann in cdr1_annotations:
        if not ann:
            _numbering[1].append( ((prev_state+1, ' '), '-') )
            prev_state += 1
        else:
            _numbering[1].append( (ann, cdr1seq[si]) )
            prev_state = ann[0]
            si += 1

    cdr2seq    = "".join([ x[1] for x in _regions[3] if x[1] != "-" ])
    cdr2length = len(cdr2seq)

    si = 0
    # Get re-annotated CDR2; CDR2 starts from 56-65. Thus range from 56 (inc.) to 66 (exc.)
    # Has a theoretical length of 10
    cdr2_annotations = get_imgt_cdr(cdr2length, 10, 56, 66)
    prev_state = 55
    for ann in cdr2_annotations:
        if not ann:
            _numbering[3].append( ((prev_state+1, ' '), '-') )
            prev_state += 1
        else:
            _numbering[3].append( (ann, cdr2seq[si]) )
            prev_state = ann[0]
            si += 1

    cdr3seq    = "".join([ x[1] for x in _regions[5] if x[1] != "-" ])
    cdr3length = len(cdr3seq)
    si = 0
    previous_state_id = 104
    for annotation in get_imgt_cdr(cdr3length, 13, 105, 118):
        if annotation is None:
            _numbering[5].append( ((previous_state_id+1, " "), "-"   ) )
            previous_state_id+=1
        else:
            _numbering[5].append( (annotation, cdr3seq[si] ) )
            previous_state_id = annotation[0]
            si+=1

    return sum(_numbering, []), startindex, endindex
  
######################
# Annotation of CDRs #
######################

def get_imgt_cdr(length, maxlength, start, end):
    """
        Symmetrically number a CDR loop (e.g. CDRL1/CDRH2 for IMGT)
        @param length:      Define the length of target CDR
        @param maxlength:   Define the theoretical limit (e.g. L1 = 12 for the IMGT scheme)
        @param start, end:  Start and end position numbers
    """
    annotations = [ None for _ in range(max(length, maxlength)) ]
    if length == 0:
        return annotations
    elif length == 1:
        annotations[0] = (start, ' ')
        return annotations

    front, back = 0, -1
    az = "ABCDEFGHIJKLMNOPQRSTUVWXYZ" 
    za = "ZYXWVUTSRQPONMLKJIHGFEDCBA"
    assert (length-maxlength) < 50, "Too many insertions for numbering scheme to handle" # We ran out of letters.

    for i in range(min(length, maxlength)):
        if i % 2:
            annotations[back] = (end + back, " ")
            back -= 1
        else:
            annotations[front] = (start + front, " ")
            front += 1

    # Add insertions around the centre point
    centrepoint = [ i for i,v in enumerate(annotations) if v == None ]
    if not centrepoint:
        return annotations

    centre_left  = annotations[min(centrepoint)-1][0] # Get the index right before the first None
    centre_right = annotations[max(centrepoint)+1][0] # Get the index right after  the first None

    # For cases with an even max length
    if not maxlength % 2:
        frontfactor, backfactor = maxlength/2, maxlength/2
    # For cases with an odd max length
    else:
        frontfactor, backfactor = (maxlength/2)+1, maxlength/2

    for i in range(max(0, length-maxlength)):
        if not i % 2:
            annotations[back] = (centre_right, za[back + backfactor])
            back -= 1
        else:
            annotations[front] = (centre_left, az[front - frontfactor])
            front += 1
    
    return annotations


def get_bc_loop(length):
    """
    Return a numbering for the BC loop in the B2 Microglobulin. It goes from 27-38 with a 2-position gap.
    """
    start, end = 27, 39 # start and end points of the BC loop. Despite having a 12-position gap, there are only 10 positions that are numbered
    annotations = [ None for _ in range(max(length, 10)) ]

    front = 0
    back  = -1
    for i in range(min(length, 10)):
        if i%2:
            annotations[back] = (end+back, " ")
            back -= 1
        else:
            annotations[front] = (start+front, " ")
            front += 1
    return annotations

def get_de_loop(length):
    """
    """
    # Control insertions which go from A-G and G-A for the DE loop
    ag = "ABCDEFG"
    ga = "GFEDCBA"
    annotations = [None for _ in range(max(length, 14))]
    back, front = -1, 0

    for i in range(length):
        if not i%2:
            annotations[back]  = (85, ga[back])
            back -= 1
        else:
            annotations[front] = (84, ag[front])
            front += 1
    return annotations

# #######
# # Aho #
# #######

# def number_aho(state_vector, sequence, chain_type):
#     """    
#     Apply the Aho numbering scheme
    
#     Rules should be implemented using two strings - the state string and the region string. 

#     There are 128 states in the HMMs. Treat X as a direct match in IMGT scheme, I is an insertion. (All X's for IMGT)

#     XXXXXXX XXX XXXXXXXXXXXXXX XXXXXXXXXXXXXXXX XXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXX XXXXX XXXX XXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXX XXXXXXXXXXX
#     AAAAAAA BBB CCCCCCCCCCCCCC DDDDDDDDDDDDDDDD EEEEEEEEEEEEEEE FFFFFFFFFFFFFFFFFFFF GGGGG HHHH IIIIIIIIIIIIIIIIIIII JJJJJJJJJJJJJ KKKKKKKKKKK


#     Regions - (N.B These do not match up with any particular definition of CDR)

#     A. 1-7 inclusive 
#     B. 8-10 inclusive. Deletion occurs at 8
#     C. 11-24 inclusive.
#     D. 25-42 inclusive (deletion surround 28) 32-42 inclusive (deletions surround 36)
#     E. 43-57 inclusive 
#     F. 58-77 inclusive (deletions surround 63). Alpha chains have deletions at 74,75
#     G. 78-82 inclusive 
#     H. 83-86 inclusive  gaps on 86 then 85
#     I. 87-106 inclusive
#     J. 107-137 inclusive gaps on 123 symetrically.
#     K. 
    
#     """
    
#     # Set up the numbering 
 
#     # State string - 'X' means the imgt position exists in the scheme. 'I' means that it should be treated as an insertion of the previous number
#     state_string =  'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'
                    
#     # Region string - regions that should be treated separately in putting the numbering together 
#     region_string =  'AAAAAAABBBCCCCCCCCCCCCCCDDDDDDDDDDDDDDDDEEEEEEEEEEEEEEEFFFFFFFFFFFFFFFFFFFFGGGGGHHHHIIIIIIIIIIIIIIIIIIIIJJJJJJJJJJJJJKKKKKKKKKKK'
# #                     0      1  2             3               4              5                   6    7   8                   9            10
#     region_index_dict = dict( zip( "ABCDEFGHIJK", range(11) ) )
    
#     # Define how the scheme's numbering differs from IMGT at the start of each region. 
#     # This is updated in the loop below
#     rels              =  {0:0, 
#                          1:0,
#                          2:0,
#                          3:0,
#                          4:2,
#                          5:2,
#                          6:2,
#                          7:2,
#                          8:2,
#                          9:2,
#                          10:21}
    
#     n_regions = 11
    
#     exclude_deletions = [1,3,4,5,7,9]    
    
#     _regions, startindex, endindex = _number_regions(sequence, state_vector, state_string , region_string,  region_index_dict, rels, n_regions, exclude_deletions)
    
#     ###############
#     # Renumbering #
#     ###############

#     _numbering = [ _regions[0], _regions[1], _regions[2],[], _regions[4], [], _regions[6], [], _regions[8],_regions[9],_regions[10] ]

#     #########################################
#     # Move the deletion around 8-10 onto 8  #
#     #########################################
#     length=len( _regions[1] )
#     if length < 3:
#         annotations = [(8," "), (9," "), (10, " ")][3-length:]
#         _numbering[1] = [ (annotations[i], _regions[1][i][1]) for i in xrange(length) ]
#     else:
#         _numbering[1] = _regions[1]


#     #########
#     # CDR 1 # - divided in two parts in the Aho scheme.
#     ######### - gaps at 28 depending on the chain type (although not absolutely defined e.g. 8fab which this will not get correct).
#     # The default number of gaps expected at 28 for each chain type is:
#     pregaps = {"H":1,"K":2,"L":1,"A":1,"B":2}[chain_type]
#     # This is a heuristic taken from the Aho paper.

#     length = len( _regions[3] )
#     # Only get rid of the default pregaps if there are insertions in the main loop.
#     insertions = max( length-16, 0)
#     for ins in xrange(max( length-16, 0)):
#         if pregaps > 0:
#             pregaps -= 1
#         else:
#             break
#     # Pre-31. 
#     positions  = [(25," "), (26," "), (27," "), (28," "), (29," "), (30," "), (31," ")]
#     if length < 7-pregaps:
#         _n = length
#     else:
#         _n = 7-pregaps
#     annotations = [ positions[i] for i in sorted([0,1,6,5,4,2,3][:_n]) ]

#     # Post 31. Delete in order 36,35,37,34,38,33,39,32,40,41,42. Insertions on 36 if present
#     _n = max( length - _n, 0 )
#     positions  = [(32," "),(33," "),(34," "),(35," "),(36," "),(37," "),(38," "),(39," "),(40," "),(41," "),(42," ")]
#     if insertions:
#         annotations += positions[:5] + [(36, alphabet[i]) for i in xrange(insertions)] + positions[5:]
#     else:
#         annotations += [ positions[i] for i in sorted([10,9,8,0,7,1,6,2,5,3,4][:_n]) ]
#     _numbering[3] = [ (annotations[i], _regions[3][i][1]) for i in xrange(length) ]

#     #########
#     # CDR 2 #
#     #########
#     # Ignore all gaps in this region and number to the aho specification.
#     # Gaps centred on 63 and also 74,75
#     # Gaps only occur on 74,75 in the alpha chain according to the aho analysis.
#     # This region describe 58 to 77 inclusive
        
#     length = len(_regions[5])
#     insertions = max( length - 20, 0 )
#     positions = [ (_, " ") for _ in xrange(58,78) ]
#     if insertions:
#         annotations = positions[:6] + [(63, alphabet[i]) for i in xrange(insertions)] + positions[6:]          
#     else:
#         deletions = 20 -length
#         if chain_type == "A":
#             removal_order = [17,16,5,4,6,3,7,2,8,1,9,0]+range(10,16)+[18,19]
#         else:
#             removal_order = [5,4,6,3,7,2,8,1,9,0]+range(10,20)
#         annotations = [ positions[i] for i in sorted( removal_order[deletions:] ) ]            
#     _numbering[5] = [ (annotations[i], _regions[5][i][1]) for i in xrange(length) ]


#     ###############################################
#     # Move deletions around 83-86 onto 85 then 86 #
#     ###############################################
#     length=len( _regions[7] )
#     if length < 4:
#         annotations = [(83," "), (84," "), (85, " "), (86, " ")][:length]
#         _numbering[7] = [ (annotations[i], _regions[7][i][1]) for i in xrange(length) ]
#     else:
#         _numbering[7] = _regions[7]

#     #########
#     # CDR 3 #
#     #########
#     # Deletions on 123. 
#     # Point of the Aho scheme is that they have accounted for all possible positions.
#     # Assumption is that no more insertions will occur.... 
#     # We'll put insertions on 123 linearly.(i.e.ABCDEF...) if they ever do.
#     # This does not happen as the HMM alignment does not recognise CDR3s that are that long.
#     length = len(_regions[9])
#     positions = [ (_, " ") for _ in xrange(107,139) ]
#     length = len(_regions[9])
#     insertions = max( length-32, 0)
#     if insertions:
#         annotations = positions[:17] + [(123, alphabet[i]) for i in xrange(insertions)] + positions[17:]
#     else:
#         deletions = 32 - length
#         removal_order = [16,17,15,18,14,19,13,20,12,21,11,22,10,23,9,24,8,25,7,26,6,27,5,28,4,29,3,30,2,31,1,0]
#         annotations = [ positions[i] for i in sorted( removal_order[deletions:] ) ]    
#     _numbering[9] = [ (annotations[i], _regions[9][i][1]) for i in xrange(length) ]

#     return _numbering[0] + \
#            _numbering[1] + \
#            _numbering[2] + \
#            _numbering[3] + \
#            _numbering[4] + \
#            _numbering[5] + \
#            _numbering[6] + \
#            _numbering[7] + \
#            _numbering[8] + \
#            _numbering[9] + \
#            _numbering[10] , startindex, endindex
#     
