#!/usr/bin/env python
"""
canonicals.py
Description: 	A simplified sequence identity based method for canonising the CDRs.

Jun 24, 2017
"""

from tcrcdr.utils import *
from tcrcdr.anarci import anarci
import numpy as np
import pickle, os, json, re

# PSSM of the above
a1_handle = open(os.path.join(os.path.dirname(__file__), "database/cdrA1_pssm.pckl"), "rb")
a2_handle = open(os.path.join(os.path.dirname(__file__), "database/cdrA2_pssm.pckl"), "rb")
a3_handle = open(os.path.join(os.path.dirname(__file__), "database/cdrA3_pssm.pckl"), "rb")
b1_handle = open(os.path.join(os.path.dirname(__file__), "database/cdrB1_pssm.pckl"), "rb")
b2_handle = open(os.path.join(os.path.dirname(__file__), "database/cdrB2_pssm.pckl"), "rb")

canonical_pssms = {
  "CDRA1": pickle.load(a1_handle),
  "CDRA2": pickle.load(a2_handle),
  "CDRA3": pickle.load(a3_handle),
  "CDRB1": pickle.load(b1_handle),
  "CDRB2": pickle.load(b2_handle)
}

# CDR boundaries are the same for B/A
IMGT_CDR = { 
                   1: list(range(27,39)), 
                   2: list(range(56,66)),
                   3: list(range(105,118)) 
           }

def get_CDRSeq(numberedseq, cdrns):
	"""
	Get the CDR sequence between the given bounds
	"""
	return ''.join([n[1] for n in numberedseq if n[0][0] in cdrns and n[1] != '-'])

def assign_canonical(sequence, cdr, return_center = True, threshold = 1):

    if cdr not in canonical_pssms:
        if return_center:
            return (None, None)
        else:
            return None

    canonical_matrices = canonical_pssms[cdr]
    cdrlen = len(sequence)

    if cdrlen not in canonical_matrices:
        if return_center:
            return (None, None)
        else:
            return None

    # For CDRA3, the threshold should be length of the loop
    if "3" in cdr:
        threshold = cdrlen

    canonicals = canonical_matrices[cdrlen]

    ## First look for identical sequence match ##
    for k,v in list(canonicals.items()):
        if len(v[2]) == 1:
            if sequence == v[2][0]:
                if return_center:
                    return k, v[1]
                else:
                    return k

    loopscores = np.array([ (k, pssm_score(sequence, v[0]), v[1]) for k,v in list(canonicals.items()) ])
    if len(loopscores) == 0:
        if return_center:
            return (None,None)
        else:
            return None

    scorearray = np.array(loopscores[:,1], float)
    
    if np.any( scorearray >= threshold):
        wheremax = np.argmax(scorearray)
        if return_center:
            return (loopscores[ wheremax ][0], loopscores[wheremax][2])
        else:
            return loopscores[ wheremax ][0]

    else:
        if return_center:
            return (None, None)
        else:
            return None

def convert_keys_to_nonbinary(dictionary):
    if isinstance(dictionary, dict):
        new_dict = {}
        for key, value in dictionary.items():
            new_key = key.decode('utf-8') if isinstance(key, bytes) else key
            new_value = convert_keys_to_nonbinary(value)
            new_dict[new_key] = new_value
        return new_dict
    elif isinstance(dictionary, list):
        return [convert_keys_to_nonbinary(item) for item in dictionary]
    else:
        return dictionary

def pssm_score(sequence, matrix):
    """
    Return the score for a given matrix
    """
    score = 0.0
    sorted_keys = sorted(matrix.keys(), key = lambda z: keysort(z, 112))

    for i,v in enumerate(sequence):
        sk_i = sorted_keys[i]

        if v.upper() in convert_keys_to_nonbinary(matrix[sk_i]):
            score += matrix[sk_i][v.encode('utf-8')]
        else:
            score += -1

    return score

def assign_fullseq(sequences):
    """
    Input a single sequence or a list/dict of sequences and get the CDR canonical forms
    """
    seqs = []
    if type(sequences) == dict:
        seqs = list(sequences.items())
    elif type(sequences) == list:
        seqs = sequences
    else: # Single sequence input
        if '/' in sequences: # must be Heavy-chain/Light-chain 
            seqh,seql = sequences.split('/')
            seqs = [('input_1',seqh),('input_2',seql)]
        else:
            seqs = [('input',args.seq)]
    
    assignresults = [{} for i in range(len(seqs))]
    annotation, alignmentdetails,_ = anarci(seqs,allow=['B','A'])

    for i in range(len(seqs)):
        seqname = seqs[i][0]
        assignresults[i].update({'seqname':seqname})
        assignresults[i].update({'input':seqs[i]})
        assignresults[i].update({'outputs':{}})
        
        # Find out if ANARCI can number this chain
        if alignmentdetails[i] == None: continue
        ct = alignmentdetails[i][0]['chain_type']
        if ct not in ["BA","AB","A","B"]: 
            print("No canonical form information for these chains %s"%chain_type)
            continue
            
        for cdrn in IMGT_CDR:
            if ct+str(cdrn) == "B3": continue
            cdrseq = get_CDRSeq(annotation[i][0][0],IMGT_CDR[cdrn])

            cluster,median = assign_canonical(cdrseq,'CDR'+ct+str(cdrn),return_center=True) # THIS IS THE FUNCTION THAT BREAKS!!!

            if cluster == None:
                assignresults[i]['outputs'][ct+str(cdrn)] = [cdrseq,'None','None']
            else:
                assignresults[i]['outputs'][ct+str(cdrn)] = [cdrseq,cluster,median]
    
    return assignresults
    
def assigncmd(args):

    seqs = []
    if os.path.isfile(args.seq):
        with open(args.seq) as f:
            _seqs = list(SimpleFastaParser(f))
        for seqid,seq in _seqs:
            if '/' in seq:
                seqh,seql = seq.split('/')
                seqh = re.sub('[\W]+','',seqh)
                seql = re.sub('[\W]+','',seql)
                seqs.append((seqid+'_1',seqh))
                seqs.append((seqid+'_2',seql))
            else:
                seqs.append((seqid,re.sub('[\W\/]+','',seq)))
    else: # Single sequence input
        if '/' in args.seq: 
            seqh,seql = args.seq.split('/')
            seqs = [('input_1',seqh),('input_2',seql)]
        else:
            seqs = [('input',args.seq)]
    
    assignresults = [{} for i in range(len(seqs))]
    annotation, alignmentdetails,_ = anarci(seqs,allow=['B','A'])
    
    for i in range(len(seqs)):
        seqname = seqs[i][0]
        assignresults[i].update({'seqname':seqname})
        assignresults[i].update({'input':seqs[i]})
        assignresults[i].update({'outputs':{}})
		  # Find if ANARCI can number this chain 
        if alignmentdetails[i] == None: continue
        ct = alignmentdetails[i][0]['chain_type']
        if ct not in ["BA","AB","A","B"]: 
            print("No canonical form information for these chains %s"%chain_type)
            continue
        for cdrn in IMGT_CDR:
            if ct+str(cdrn) == "B3": continue
            cdrseq = get_CDRSeq(annotation[i][0][0],IMGT_CDR[cdrn])
            cluster,median = assign_canonical(cdrseq,'CDR'+ct+str(cdrn),return_center=True)
            if cluster == None:
                assignresults[i]['outputs'][ct+str(cdrn)] = [cdrseq,'None','None']
            else:
                assignresults[i]['outputs'][ct+str(cdrn)] = [cdrseq,cluster,median]
    
    # Output format
    if args.outputdir == False:
        print_result(assignresults)
    elif args.outputdir == '':
        pass # No output
    elif args.outputformat in ['txt','csv']:
        opf = write_txt(assignresults, args.outputdir)
        print(('Write results to %s' % opf))
    elif args.outputformat =='json':
        opf = os.path.join(args.outputdir,'{0}.{1}'.format(time.time(),args.outputformat))
        if not os.path.exists(args.outputdir): os.mkdir(args.outputdir) 
        with open(opf,'wb') as f:
            json.dump(assignresults,f)
        print(('Write results to %s' % opf))
    return assignresults
